/****************************** Module Header *******************************
*
* Module Name: edit.e
*
* Copyright (c) Netlabs EPM Distribution Project 2002
*
* $Id: edit.e 3480 2019-02-28 18:26:17Z aschn $
*
* ===========================================================================
*
* This file is part of the Netlabs EPM Distribution package and is free
* software.  You can redistribute it and/or modify it under the terms of the
* GNU General Public License as published by the Free Software
* Foundation, in version 2 as it comes in the "COPYING" file of the
* Netlabs EPM Distribution.  This library is distributed in the hope that it
* will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
****************************************************************************/

; Definitions for editing (loading) files.
; See also: SLNOHOST.E/SAVELOAD.E/E3EMUL.E and FILE.E.

; ---------------------------------------------------------------------------
; EPM bug during heavy processing, e.g. file loading:
;
; When too many procs are nested, esp. when changing between C and E code
; permanently, sometimes E or C code will not be executed.
;
; That was first seen when the call of NepmdActivateHighlight suppressed the
; usually following select event after all files are loaded.
; (Instead of letting the C function NepmdActivateHighlight execute the
; Highlight command, now NepmdQueryHighlightArgs is used to retrieve
; only the args for Highlight. Highlight is executed from E code now.)
;
; To workaround that bug, one has to change the code processing:
;
;    o  Try to avoid too many nested procs.
;
;    o  Try to avoid executing E commands from C code.
;
;    o  Split the E code before the critical point to an additional defc.
;       An additional defproc won't help here. Apparently the previous code
;       is always completed, when a new command is executed, at least most
;       of the times. To ensure that any previous code is completed, the
;       additional command must be executed with postme.

; ---------------------------------------------------------------------------
; Slow processing caused by specific code:
;
; The file load processing can easily be slowed down. It is extremely
; vulnerable to following code:
;
;    o   sayerror statement or command
;
;    o   postme command or other posted window messages.
;
; Therefore these should be avoided during file loading, if possible.

; ---------------------------------------------------------------------------
; Unreliable sayerrors:
;
; During heavy processing, sayerror statements are often gobbled, appear in
; the wrong order or even doubled in the message list. Therefore better use
; the NepmdPmPrintf proc or the dprintf proc, together with a PmPrintf
; application for debugging.

; ---------------------------------------------------------------------------
; Pre-processing before calling the LoadFile defproc. Same syntax as
; LoadFile. Calls itself LoadFile at the end. (LoadFile is defined in
; SLNOHOST.E / SAVELOAD.E / E3EMUL.E. Without HOST_SUPPORT, always SLNOHOST.E is
; included.)
; Checks a filespec, even wildcards, to remove REXX EAs before loading.
; Called by defc Edit.
defproc PreLoadFile( FileSpec, Options)
   universal filestoload
   universal filestoloadmax  -- still used for 'xcom e' and ProcessFilesLoaded

   if filestoload = '' then
      filestoload = 0
   endif
   if filestoloadmax = '' then
      filestoloadmax = 0
   endif

   StrippedFileSpec = FileSpec
   if leftstr( FileSpec, 1) = '"' & rightstr( FileSpec, 1) = '"' then
      StrippedFileSpec = substr( FileSpec, 2, length( FileSpec) - 2)
   endif

   -- Experimental URL support, configurable via URL_LOAD_TEMPLATE.
   -- Default is to use wget for downloading it.
   if pos( '://', StrippedFileSpec) > 0 then
      call LoadUrl( StrippedFileSpec, Options)
      -- Don't execute the rest of PreLoadFile and Edit -> LoadFile
      stop
   endif

   -- Change slashes to backslashes, except for URLs
   if pos( '://', StrippedFileSpec) = 0 then
      StrippedFileSpec = translate( StrippedFileSpec, '\', '/')
   endif

   fWildcard = (pos( '*', StrippedFileSpec) + pos( '?', StrippedFileSpec) > 0)

   -- Resolve wildcards in FileSpec to delete REXX EAs for every REXX file
   -- and for setting universal vars
   Handle   = ''  -- always create a new handle!
   Filename = ''
   fNoMoreFiles = 0
   do while fNoMoreFiles = 0
      if fWildcard then
         -- If FileSpec contains wildcards then find Filenames
         if NepmdGetNextFile( StrippedFileSpec, Handle, Filename) then
            NepmdGetNextClose( Handle)
            filestoload = filestoload + 1
         else
            fNoMoreFiles = 1
            leave
         endif
      else
         -- If FileSpec doesn't contain wildcards then set Filename to make the
         -- edit command add a file to the ring, even when it doesn't exist
         Filename = StrippedFileSpec
         filestoload = 1
         fNoMoreFiles = 1
      endif
      --sayerror 'StrippedFileSpec = 'StrippedFileSpec', Filename = 'Filename
      dprintf( 'EDIT', 'PreLoadFile: StrippedFileSpec = 'StrippedFileSpec', Filename = 'Filename)

      -- Delete REXX EAs if extension is found in RexxEaExtensions.
      call PreloadDeleteRexxEas( Filename)

      -- GetMode doesn't work here, because it tries to write the fid'.mode'
      -- array var. At this time the file is not loaded, so the fileid is not set.
      -- But calling NepmdQueryDefaultMode( Filename) would work.
      -- We should better determine the mode here, save it in a array var with the
      -- .filename as identifier, and replace the identifier with the fileid later
      -- at defload (or replace the array var with the final one).

compile if 0
      -- Experimentell codepage support
      call PreloadProcessCodepage( Filename)
compile endif

   enddo
   filestoloadmax = filestoload

   rc = 0
   return

; ---------------------------------------------------------------------------
const
compile if not defined( REXX_EA_EXTENSIONS)
   -- Extensions, whose REXX EAs shall be removed before loading
   -- to workaround the EPM bug that it can't load EAs > 32KB.
   REXX_EA_EXTENSIONS = 'CMD ERX'
compile endif

; Delete REXX EAs if extension is found in RexxEaExtensions.
; Use the extension here instead of the mode to avoid determining the
; mode twice: here and at defload.
; Note: The use of array vars containing the fileid to become file-specific
; does only work properly at or after defload. Therefore the mode should be
; determined at defload.
; Syntax: PreloadDeleteRexxEas( Filename)
defproc PreloadDeleteRexxEas( Filename)
   rc = 0
   do once = 1 to 1
      p = lastpos( '\', Filename)
      Name = substr( Filename, p + 1)  -- strip path
      p = lastpos( '.', Name)
      if p <= 1 then
         leave
      endif
      Ext = translate( substr( Name, p + 1))
      if not wordpos( Ext, REXX_EA_EXTENSIONS) then
         leave
      endif
compile if 0
      -- Calling here another proc doesn't always work reliable.
      if GetReadOnly( Filename) <> 0 then
         leave
      endif
compile else
      -- Inserting the code here directly always works.
      fReadonly = 0
      Attr = NepmdQueryPathInfo( Filename, 'ATTR')
      if Attr = '' then  -- file doesn't exist
         --sayerror 'Attributes for "'Filename'" can''t be retrieved, rc = 'rc
         return
      elseif length( Attr) = 5 then
         fReadonly = (substr( Attr, 5, 1) = 'R')
      endif
      -- dprintf( 'PreloadDeleteRexxEas', Filename' - 'attr', fReadonly = 'fReadonly)
      if fReadonly then
         leave
      endif
compile endif
      --sayerror 'Removing REXX EAs with NepmdLib from 'Filename
      rc = 0
      NepmdDeleteRexxEa( Filename)
   enddo
   return

; ---------------------------------------------------------------------------
; Experimental codepage support.
; The .codepage field var exists, but is not yet used.
defproc PreloadProcessCodepage( Filename)
   rc = 0
   do once = 1 to 1
      CurCodepage = 'cp850'
      Codepage = NepmdQueryStringEa( Filename, 'EPM.CODEPAGE')
      if rc then
         leave
      endif
      if Codepage = '' then
         leave
      endif

      Codepage = upcase( Codepage)
      if Codepage = 'CP1004' then
         Codepage = 'latin-1'
      endif

      -- Note: GNU Recode deletes all EAs
      lp = lastpos( '.', Filename)
      if lp > 1 then
         EaFile = leftstr( Filename, lp)'ea'
      else
         EaFile = Filename'.ea'
      endif
      StripEaCmd = 'eautil' Filename EaFile '/s /r'
      --JoinEaCmd  = 'eautil' Filename EaFile '/j /m'
      JoinEaCmd  = 'eautil' Filename EaFile '/j /o'
      RecodeCmd  = 'recode 'Codepage':'CurCodepage Filename

      -- Executing all in one cmd doesn't work
/*
      Cmd = StripEaCmd'^&'RecodeCmd'^&'JoinEaCmd
      quietshell Cmd
*/
      quietshell StripEaCmd
      quietshell RecodeCmd
      quietshell JoinEaCmd

      -- Delete EA, because it will not be reset on save yet
      NepmdDeleteStringEa( Filename, 'EPM.CODEPAGE')
   enddo
   return

; ---------------------------------------------------------------------------
defc LoadDir
   if arg( 1) = '' then
      Dir = directory()
   endif
   call LoadDir( Dir)

defproc LoadDir( FileSpec)
   rc = 3  -- Path not found
   do once = 1 to 1
      if leftstr( FileSpec, 1) = '"' & rightstr( FileSpec, 1) = '"' then
         StrippedFileSpec = substr( FileSpec, 2, length( FileSpec) - 2)
      endif
      if not NepmdDirExists( StrippedFileSpec) then
         -- NepmdDirExists sets rc = 0 also for existing files, probably from
         -- DosQueryPathInfo. Therefore better check the returned value.
         rc = 3  -- Path not found
         leave
      endif

      fOpenDir = 1
      -- For Shell buffers: enter current dir name
      if IsAShell( .filename) then
         ShellDir = ''
         ShellCmd = ''
         call ShellParsePromptLine( ShellDir, ShellCmd)
         if length( ShellDir) > 0 then
            fOpenDir = 0
            if ShellCmd = '' then
               endline
               NewCmd = 'cdd' FileSpec
            else
               NewCmd = FileSpec
            endif
            keyin NewCmd
         endif
      endif

      -- If not a shell: open dir listing
      if fOpenDir then
         'Dir 'FileSpec
      endif
      rc = 0
   enddo
   return

; ---------------------------------------------------------------------------
;compile if LINK_HOST_SUPPORT & (HOST_SUPPORT='EMUL' | HOST_SUPPORT='E3EMUL')
; compile if not defined( MVS)
;    MVS = 0
; compile endif
; compile if not defined( E3MVS)
;    E3MVS = 0
; compile endif
; compile if not defined( HOST_LT_REQUIRED)
;    HOST_LT_REQUIRED = 0
; compile endif
;compile endif

; Returns 1 if not a VM file, otherwise 0. Sets rest.
defproc ParseHostFileSpec( var rest)
   fIsVmFile = 0
   do once = 1 to 1
compile if HOST_SUPPORT & not SMALL
      p  = length( rest) + 1  -- If no delimiters, take to the end.
      p1 = pos( ',', rest)
      if not p1 then
         p1 = p
      endif
      p2 = pos( '/', rest)
      if not p2 then
         p2 = p
      endif
      p3 = pos( '"', rest)
      if not p3 then
         p3 = p
      endif
      p4 = pos( "'", rest)
      if not p4 then
         p4 = p
      endif
      p5 = pos( '[', rest)
      if not p5 then  -- Allow for [FTO]
         p5 = p
      endif
 compile if HOST_SUPPORT='EMUL' | HOST_SUPPORT='E3EMUL'
  compile if MVS
      p4 = p     -- Can't use single quote for commands if allowing MVS files
  compile endif
 compile else
      p5 = p
 compile endif
      p = Min( p1, p2, p3, p4, p5)
      FileSpec = leftstr( rest, p - 1)
      if VMfile( FileSpec, more) then    -- tricky - VMfile modifies file
         if p = p1 then                  -- Keep any except comma in string
            p = p + 1
         endif
         rest = more substr( rest, p)
         fIsVmFile = 1
      endif
compile endif
   enddo
   return (not fIsVmFile)

; Sets universal fto. Sets rest.
defproc ParseHostFto( var rest)
compile if (HOST_SUPPORT = 'EMUL' | HOST_SUPPORT = 'E3EMUL') & not SMALL
   universal fto              -- Need this passed to loadfile

   if substr( strip( rest, 'L'), 1, 1) = '[' then
      parse value rest with '[' fto ']' rest
   else
      fto = ''                --  reset for each file
   endif
compile endif
   return

; MVS file specs can start with '. Returns flag if single quote should be
; parsed to get 'cmd'.
defproc ParseEditCmd
   fParseCmd = 1
compile if (HOST_SUPPORT = 'EMUL' | HOST_SUPPORT = 'E3EMUL') & not SMALL
 compile if MVS & not HOST_LT_REQUIRED  -- MVS file specs can start with '
   fParseCmd = 0
 compile endif
compile endif
   return fParseCmd

; ---------------------------------------------------------------------------
;
; epm /r filename     tries to find the file in the current dir of
;                     the starting environment. Apparently
;                     parse_filename is executed before the directory
;                     is overtaken from the topmost EPM window.
; epm /r 'e filename' tries to find the file in the current dir of
;                     the topmost existing EPM window. The 'e' command
;                     is executed after the main initialization is
;                     done.
; A possible fix to make both work the same: Move parameters to an additional
; cmd and execute it from the edit cmd.
;
; When multiple EPM cmds should be executed by separate epm /r 'cmd' calls,
; a short pause (e.g. 0.2 s) before each epm /r call is required.
; A possible fix to make condecutive calls work without a pause: Put
; submitted cmds on a stack and process the next one only if the last one
; is done.

; ---------------------------------------------------------------------------
; This DEFC EDIT eventually calls the built-in edit command (xcom edit), by
; calling PreLoadFile -> LoadFile, that does additional processing before
; the file(s) is/are loaded.
;
; The EDIT command can not only be executed from EPM's commandline, but is
; also prepended to the submitted args when EPM is started. Before that,
; EPM removes the options that it knows. The rest of the arg string is always
; submitted to the EDIT command. (Main arg processing is done in MAIN.E.)
;
; Parse off each file individually.  Files can optionally be followed by one
; or more commands, each in quotes.  The first file that follows a host file
; must be separated by a comma, an option, or a (possibly null) command.
;
; EPM doesn't give error messages from XCOM EDIT, so we have to handle that
; for it.

defc e, ed, Edit, epm
   universal mainstate
   universal default_edit_options
   universal cureditcmd
   universal firstloadedfid  -- first file for this edit cmd, set here
   universal firstinringfid  -- first file in the ring, set by defmain
                             -- Both values are set by defmain, if only
                             -- the unknown file was loaded (via xcom edit).

   edit_rc = ''

   do once = 1 to 1
      getfileid startfid  -- save fid of topmost file before current edit cmd

      -- Workaround: This avoids unwanted window scrolling of the previous file.
      'VSyncCursor'

      --call ResetHiliteModeList()

      -- Set current edit cmd to let other commands differ between several ways
      -- of file loading.
      -- Other commands that execute 'Edit', can set this universal var before
      -- and then it will not be overwritten by 'Edit'. ProcessFilesLoaded
      -- will reset it.
      -- This is currently used for RestorePos and RingSavePos.
      if cureditcmd = '' then
         cureditcmd = 'EDIT'
      endif

      args = strip( arg( 1))

      if args = '' then  -- The original 'edit' by itself goes to next file
         --nextfile      -- Removed to make 'epm /r' only bring an EPM window to the foreground
                         -- instead of switching to the next file in the ring.
         leave
      endif

      Options = default_edit_options

      NumErrors = 0
      FirstFid  = ''

      fParseCmd = 1
      if isadefproc( 'ParseEditCmd') then
         fParseCmd = ParseEditCmd()  -- MVS file specs can start with '
      endif

      rest = args
      do while rest <> ''
         rest = strip( rest, 'L')
         -- Remove leading ',' from rest
         -- A ',' separates multiple <FileSpec> '<command>' segments.
         if substr( rest, 1, 1) = ',' then
            rest = strip( substr( rest, 2), 'L')
         endif
         -- Get first char
         ch = substr( rest, 1, 1)

         if ch = "'" & fParseCmd then
            parse value rest with (ch) Cmd (ch) rest
            do while substr( rest, 1, 1) = ch & pos( ch, rest, 2)
               parse value rest with (ch) p (ch) rest
               Cmd = Cmd''ch''p
            enddo
            CurEditCmd = Cmd  -- set universal var to determine later in LOAD.E if pos shall be restored from EA

            -- Recognize a prepended PostMe. If it exists, Cmd2 contains the rest.
            Cmd2 = ''
            if Cmd = '' then
               -- nop
            elseif upcase( word( Cmd, 1)) = 'POSTME' then
               Cmd2 = subword( Cmd, 2)
            endif

            -- Execute Cmd if specified.
            -- Prepend GoTo if a line number was specified to minimize scrolling.
            if Cmd = '' then
               -- nop
            elseif IsNum( Cmd) then
               'GoTo' Cmd
            elseif IsNum( Cmd2) then
               'PostMe GoTo' Cmd2
            else
               Cmd
            endif

         elseif ch = '/' then       -- Option
            parse value rest with opt rest
            Options = Options upcase( opt)

         else
            -- Remove double quotes
            if ch = '"' then
               p = pos( '"', rest, 2)
               if p then
                  FileSpec = substr( rest, 1, p)
                  rest = substr( rest, p + 1)
               else
                  sayerror INVALID_FILENAME__MSG
                  rc = -275  -- MISSING_FILENAME_RC
                  return
               endif
            else
               fParseFileSpec = 1
               if isadefproc( 'ParseHostFileSpec') then
                  fParseFileSpec = ParseHostFileSpec( rest)  -- Returns 1 if not a VM file
               endif
               if fParseFileSpec then
                  -- Recogniation of ',' as segment parser is changed from ','
                  -- to ', ' (at least 1 space after the comma) in order to
                  -- support filenames like: "filename.ext,v", as created by CVS.
                  -- Parse at next ', '. Removes ', ' from FileSpec (DOS and VM)
                  -- A ', ' separates multiple <FileSpec> '<command>' segments.
                  parse value rest with FileSpec rest2
                  if pos( ', ', FileSpec) then
                     parse value rest with FileSpec ', ' rest
                  else
                     rest = rest2
                  endif
               endif
               if isadefproc( 'ParseHostFto') then
                  call ParseHostFto( rest)
               endif
            endif  -- ch = '"' else (Remove doublequotes)

            call Parse_Filename( FileSpec, .filename)

            -- Enquote FileSpec if required
            if (not (leftstr( FileSpec, 1) = '"' & rightstr( FileSpec, 1) = '"')) &
               verify( FileSpec, " &|<>',", 'M') then
               FileSpec = '"'FileSpec'"'
            endif

            -- Get real case of file system object (returns the input if it doesn't exist)
            FileSpec = GetRealCase( FileSpec)

            -- Handle dirs specially
            call LoadDir( FileSpec)
            if not rc then  -- on success
               iterate
            endif

            -- Preload the file
            call PreLoadFile( FileSpec, Options)
            if rc then
               dprintf( 'rc from PreLoadFile( FileSpec) = 'rc)
               -- Don't process LoadFile when LoadUrl returned rc <> 0
               iterate
            endif

            -- Load the file
            call LoadFile( FileSpec, Options)  -- 'xcom e' sets rc
            edit_rc = rc  -- restore this rc at the end

            if edit_rc then
               dprintf( sayerrortext( edit_rc)': 'FileSpec)
               sayerror sayerrortext( edit_rc)': 'FileSpec
               NumErrors = NumErrors + 1
            else
               if FirstFid = '' then
                  -- If rc = 0, then set FirstFid:
                  getfileid FirstFid
               endif
            endif
         endif  -- ch = ... (not "cmd")

      enddo  -- while rest <> ''

      if NumErrors > 1 then
         MessageNWait( MULTIPLE_ERRORS__MSG)
      endif

      dprintf( 'EDIT', 'arg( 1) = ['arg( 1)'], FirstFid = ['FirstFid'], ['FirstFid.filename']')
      -- If 1 or more files are loaded by the current edit cmd (or if loadfile has returned rc = 0):
      if FirstFid <> '' then

         -- activatefile is now executed in ProcessFilesLoaded with postme.
         -- This finally works properly. With activatefile here the ring would get messed.
         --activatefile FirstFid
         -- Set fid for ProcessFilesLoaded:
         firstloadedfid = FirstFid

         -- Initialize firstinringfid if not already set by a previous edit command:
         if firstinringfid = '' then
            firstinringfid = firstloadedfid
         endif
      endif

      if upcase( cureditcmd) <> 'RESTORERING' then
         'AtStartup AddToHistory EDIT' args
      endif

   enddo

   if mainstate <> 1 & mainstate <> 2 then
      'PostMe PostMe ProcessSelect'
      mainstate = 1
   endif

   if edit_rc <> '' then
      rc = edit_rc
   endif

; ---------------------------------------------------------------------------
;  New in EPM.  Edits a file in a different PM window.  This means invoking
;  a completely new instance of E.DLL, with its own window and data.  We do it
;  by posting a message to the executive, the top-level E application.
defc o, Open
   fname = strip( arg( 1))                  -- Remove excess spaces
   call Parse_Filename( fname, .filename)   -- Resolve '=', if any

   call windowmessage( 0, getpminfo( APP_HANDLE),
                       5386,                   -- EPM_EDIT_NEWFILE
                       Put_In_Buffer( fname),
                       1)                      -- Tell EPM to free the buffer.

; ---------------------------------------------------------------------------
; LAM - Edit a file along the EPATH.  This command will be included if
; the user is including the required routines.  If you've done a
;   SET EPMPATH=d:\EPM;d:\my_emacs;d:\E_macros
; and then do
;   ep stdcmds.e
; that will load this file, just as if you had entered
;   e d:\e_macros\stdcmds.e
; 1994/03/02:  Can specify '.' as the path to use the default, and can
;              include additional arguments for the E command.  E.g.,
;                 ep mystuff.e . '/spell'
defc ep, EPath
   parse arg Filename Pathname Rest
   if Pathname = '' | Pathname = '.' then
      if FileType( Filename) = 'CMD' then
         PathnameList = 'PATH'
      else
         PathnameList = EPATH 'PATH'
      endif
   else
      PathnameList = Pathname
   endif
   do w = 1 to words( PathnameList)
      Pathname = word( PathnameList, w)
      call Parse_Filename( Filename,
                           substr( .filename, lastpos( '\', .filename) + 1))
      findfile Newfile, Filename, Pathname  -- find Filename in Pathname env var
      if not rc then  -- if found
         leave
      endif
   enddo
   if rc then  -- use specified Filename if not found
      Newfile = Filename
   endif
   Cmd = strip( 'e' Newfile Rest)
   Cmd

; ---------------------------------------------------------------------------
defc op, OPath, OpenPath
   "Open 'ep "arg( 1)"'"

; ---------------------------------------------------------------------------
defc New
   getfileid startfid
   'xcom e /n'
   if rc <> -282 then  -- sayerror 'New file'
      return
   endif
   getfileid newfid
   activatefile startfid
   temp = startfid  -- temp fix for some bug
   'Quit'
   getfileid curfid
   activatefile newfid
   if curfid = startfid then  -- Wasn't quit; user must have said Cancel to Quit dlg
      'xcom Quit'
   endif

; ---------------------------------------------------------------------------
; Move current file to a newly opened EPM window.
defc NewWindow
   if leftstr( .filename, 5) = '.DOS ' then
      fn = "'"substr( .filename, 6)"'"

   elseif .filename = '.tree' then
      parse value .titletext with Cmd ': ' Args
      fn = "'"Cmd Args"'"

   elseif IsAShell() then
      epmdir = directory()

      call pSave_Pos( savedpos)
      getsearch oldsearch
      -- search (reverse) in command shell buffer for the prompt and
      -- retrieve the current directory
      -- goto previous prompt line
      ret = ShellGotoNextPrompt( 'P')
      CurDir = ''
      Cmd = ''
      if not ret then
         call ShellParsePromptLine( CurDir, Cmd)
      endif
      ShellCmd = 'Shell'
      if CurDir <> '' then
         ShellCmd = ShellCmd 'cdd' CurDir
      endif
      setsearch oldsearch
      call pRestore_Pos( savedpos)

      fn = "'mc +cd "epmdir"+"shellcmd"'"

   elseif leftstr( .filename, 1 ) = '.' then  -- other temp file
      fn = ''

   else
      if .modify then
         'Save'
         if rc then
            sayerror ERROR_SAVING_HALT__MSG
            return
         endif
      endif
      fn = .filename
      if fn = GetUnnamedFileName() then
         fn = ''
      elseif pos( ' ', fn) then  -- support for spaces in filename
         fn = '"'fn'"'
      endif
      if .readonly then
         fn = '/r' fn
      endif
   endif

   PostMcArg = ''
   if fn <> '' then
      call pSave_Pos( savedpos)
      -- add arg for 'mc'
      PostMcArg = ';RestorePos 'savedpos
   endif
   -- Valid EPM commandline options are always regnized, even when they
   -- are enclosed in '...' to signalize the submitted command part.
   if GetMode() = 'BIN' then
      -- prepend 'binedit'
      fn = "'BinEdit "fn"'"
   endif
   if PostMcArg > '' then
      "Open" fn "'PostMe mc "PostMcArg"'"
   else
      'Open' fn
   endif
   'Quit'

; ---------------------------------------------------------------------------
; Edit binary files in EPM.
; linebreak after 64 chars
;    Program object:
;       Name      : EPM Bin
;       Parameters: 'binedit %*'
defc be, BinEdit
;   universal default_save_options
   -- Change default options for save command to simply save a bin file
   -- with F2. Therefore this command should be used in a separate EPM
   -- window only.
;   default_save_options = '/ne /ns /nt'
   'HookAdd loadonce tabs 1 noea'
   'HookAdd loadonce tabkey on'
   'HookAdd loadonce matchtab off'
   'HookAdd loadonce mode bin noea'
   'e /t /64 /bin "'arg( 1)'"'    -- options should go before filename                    <-- Todo: parse options and filename
                                 -- /64 doesn't work if run from a program object.
   'postme AvoidSaveOptions /e /s /t /o /l'
;   if insert_state() then
;      -- switch to overwrite mode
;      insert_toggle
;   endif

; ---------------------------------------------------------------------------
; linebreak at lineend chars
defc ble, BinLineEdit
;   universal default_save_options
   -- Change default options for save command to simply save a bin file
   -- with F2. Therefore this command should be used in a separate EPM
   -- window only.
;   default_save_options = '/ne /ns /nt'
   'HookAdd loadonce tabs 1 noea'
   'HookAdd loadonce tabkey on'
   'HookAdd loadonce matchtab off'
   'HookAdd loadonce mode bin noea'
   'e /t /bin "'arg( 1)'"'        -- options should go before filename
   'postme AvoidSaveOptions /e /s /t /o /l'
;   if insert_state() then
;      -- switch to overwrite mode
;      insert_toggle
;   endif

; ---------------------------------------------------------------------------
; linebreak at maxmargin
defc bme, BinMaxEdit
;   universal default_save_options
   -- Change default options for save command to simply save a bin file
   -- with F2. Therefore this command should be used in a separate EPM
   -- window only.
;   default_save_options = '/ne /ns /nt'
   'HookAdd loadonce tabs 1 noea'
   'HookAdd loadonce tabkey on'
   'HookAdd loadonce matchtab off'
   'HookAdd loadonce mode bin noea'
   'e /t /1599 /bin "'arg( 1)'"'  -- options should go before filename
   'postme AvoidSaveOptions /e /s /t /o /l'
;   if insert_state() then
;      -- switch to overwrite mode
;      insert_toggle
;   endif

; ---------------------------------------------------------------------------
; Open a file dialog to select a file for binary editing in a new edit
; window. No args.
defc OpenBinDlg
   filemask = '*.exe;*.dll'
   title    = 'Select a binary file'
   "o 'filedlg "title", be, "filemask"'"  -- works, but Cancel opens an empty window
   --"filedlg "title", o 'be', "filemask
   -- A command that returns the hwnd of the newly opened window is needed.

; ---------------------------------------------------------------------------
; Finds EPM macro files <basename>.e in Dir of arg( 1) and EPMMACROPATH.
; Opens listbox to select one file if multiple found.
; <basename> is parsed from arg( 1), so arg( 1) may have any extension.
; For association with *.ex (type: EX file) and *.e:
;    Program object:
;       Name      : EPM Edit macro file
;       Parameters: /r 'editmacrofile %*'
; From command line:
;    start epm /r 'editmacrofile <basename>'
; Todo?: handle UNC (\\server\resource) names
defc EditMacroFile

   PathName = 'EPMMACROPATH'
   SearchInDirOfFile = 1

   File = strip( arg( 1))
   if File = '' then
      return
   endif
   -- Remove ".."
   len = length( File)
   if substr( File, 1, 1) = '"' & substr( File, len, 1) = '"' then
      File = substr( File, 2, len - 2)
   endif
   if File = '' then
      return
   endif

   -- Get full pathname if exists
   File = NepmdQueryFullName( File)

   -- Build MacroFileName
   lp1 = lastpos( '\', File)
   Name = substr( File, lp1 + 1)
   Dir = substr( File, 1, Max( 0, lp1 - 1))
   lp2 = lastpos( '.', Name)
   if lp2 = 0 | lp2 = 1 then
      BaseName = Name
   else
      BaseName = substr( Name, 1, Max( 0, lp2 - 1))
   endif
   MacroFileName = BaseName'.e'

   Found = 0
   FoundOnlyInDirOfFile = 0
   MacroFileList = ''
   LastMacroFile = ''
   Len = 0

   -- Search MacroFile in dir of File first
   if SearchInDirOfFile = 1 then
      MacroFile = Dir'\'MacroFileName
      if NepmdFileExists( MacroFile) then
         Found = Found + 1
         Len = Max( Len, length( MacroFile))
         FoundOnlyInDirOfFile = 1
         MacroFileList = MacroFileList''\1''MacroFile
         LastMacroFile = MacroFile
      endif
   endif

   -- Get value of PathName
   Path = Get_Env( PathName)
   rest = Path
   -- Search in all parts of Path
   do while rest <> ''
      parse value rest with NextPath';'rest
      -- Search in NextPath
      MacroFile = NextPath'\'MacroFileName
      if NepmdFileExists( MacroFile) then
         if upcase( MacroFile) = upcase( LastMacroFile) then
            FoundOnlyInDirOfFile = 0
            iterate
         endif
         Found = Found + 1
         Len = Max( Len, length( MacroFile))
         MacroFileList = MacroFileList''\1''MacroFile
         LastMacroFile = MacroFile
      endif
   enddo

   if Found = 0 then
      'postme sayerror "'MacroFileName'" not found in directory of "'Name'" or in 'PathName
   elseif Found = 1 then
      -- Use postme here to delay the msg until all other msgs are processed.
      -- Otherwise the last msg would be 'Link completed, module # ?' if a new window
      -- is opened. But sometimes it doesn't work though.
      if FoundOnlyInDirOfFile = 1 then
         'postme sayerror Found macro file "'MacroFileName'" only in directory of "'Name'"'
      else
         'postme sayerror Found 1 macro file "'MacroFileName'" in 'PathName
      endif
      parse value MacroFileList with \1''select
      'e' select
   else
      -- Multiple files found, open listbox
      Title = 'Select macro file 'MacroFileName
         -- Text to appear in 1 line below the title, above the list
         -- Only 1 line (no \13 or \n) allowed in Text, but \9 is recognized.
      if FoundOnlyInDirOfFile = 1 then
         Text = 'Multiple macro files found in directory of "'Name'" and 'PathName'.'
      else
         Text = 'Multiple macro files found in 'PathName'.'
      endif
      TextZ = '  'Text\0    -- add 2 spaces for proper left alignment of text and zero termination
      ItemList = MacroFileList             -- first char is separator
      ButtonList = '/~Edit/~Open/Cancel'   -- first char is separator, first button is selected
      Selection = 1                        -- selected item of ItemList
      Height = Min( Found, 12)             -- make the list max. 12 lines high
      Width  = Max( Len, 50)               -- make the list min. 50 ??? wide
      refresh  -- Add everytime a refresh before opening a listbox.
               -- Otherwise sometimes only a part of it is shown.
      select = ListBox( Title,
                        ItemList,
                        ButtonList,        -- buttons
                        0, 0,              -- top, left,
                        Height, Width,     -- height, width,
                        GethWndC( APP_HANDLE) || atoi( Selection) || atoi( 1) || atoi( 0) ||
                        TextZ)
      parse value select with button 2 select \0  -- get button and (select = selected item)
      select = strip( select, 'B', \1)  -- sometimes the returned value for Cancel is \1
      if select = '' then               -- Cancel ==> no item is returned
         return
      elseif button = \1 then           -- Button1
         'e' select
      elseif button = \2 then           -- Button2
         'o' select
      endif
   endif
   return

; ---------------------------------------------------------------------------
; Edits a user macro file. If file doesn't exist, the contents of the
; Netlabs file is copied to the new user file buffer. A file is only created
; if a user saves the buffer afterwards.
defc EditCreateUserMacro
   MacroFile = arg( 1)
   lp = lastpos( '\', MacroFile)
   MacroFileName = substr( MacroFile, lp + 1)
   'EditCreateUserFile macros\'MacroFileName

; ---------------------------------------------------------------------------
; Edits a user macro file. If file doesn't exist, the contents of the
; Netlabs file is copied to the new user file buffer. A file is only created
; if a user saves the buffer afterwards.
; Syntax: EditCreateUserFile <subpath>\<filename>, e.g.: bin\epm.env
defc EditCreateUserFile
   MacroFile = strip( arg( 1))
   UserDir     = Get_Env( 'NEPMD_USERDIR')
   NetlabsDir  = Get_Env( 'NEPMD_ROOTDIR')'\netlabs'
   UserFile    = UserDir'\'MacroFile
   NetlabsFile = NetlabsDir'\'MacroFile
   if Exist( UserFile) then
      'Edit 'UserFile
   else
      'Edit 'NetlabsFile
      if not rc then
         .filename = UserFile
         -- Disabled. The header could contain useful docs. In some cases
         -- also the RCS id may be helpful to determine the original version
         -- of the netlabs file.
         --'postme mc ;1;RemoveNextHeader'

         'postme postme EditCreateUserFileMsgBox' UserFile
      endif
   endif

; ---------------------------------------------------------------------------
defc EditCreateUserFileMsgBox
   NewFile = strip( arg( 1))
   NewFilename = ''
   if NewFile <> '' then
      lp = lastpos( '\', NewFile)
      NewFilename = substr( NewFile, lp + 1)
   endif
   Mode = GetMode()
   Title = 'New user file: 'NewFilename
   Text = ''
   Text = Text || 'The current file was copied from the NETLABS tree.'
   Text = Text || ' On saving, it will be created in the user tree and'
   Text = Text || ' will override the original NETLABS file.'
   if Mode = 'E' then
      Text = Text || \n\n
      Text = Text || 'Note that a macro file needs recompilation before'
      Text = Text || ' it gets applied. In that case, execute RecompileNew'
      Text = Text || ' via Run -> Macros -> "Recompile new user macros".'
   endif
   Style = MB_OK + MB_INFORMATION + MB_MOVEABLE
   ret = WinMessageBox( Title,
                        Text,
                        Style)

; ---------------------------------------------------------------------------
; Remove next comment block from a buffer, starting at cursor line.
defc RemoveNextHeader
   saved_autosave = .autosave
   .autosave = 0
   saved_modify = .modify
   call NextCmdAltersText()
   InfolineRefresh = 0

   cComment = 0
   fHeader = 0
   -- Check comment char on current line
   LineCommentChars = '; : # %'
   CommentChars = ''
   getline line
   ch1 = leftstr( line, 1)
   ch2 = leftstr( line, 2)
   wp = wordpos( ch1, LineCommentChars)
   if ch2 = '/*' then
      CommentChars = '/* */'
      fHeader = 1
   elseif wp > 0 then
      CommentChars = word( LineCommentChars, wp)
      fHeader = 1
   endif

   do while fHeader
      getline line

      if words( CommentChars) = 2 then
         w1 = word( CommentChars, 1)
         w2 = word( CommentChars, 2)
         if leftstr( line, length( w1)) = w1 then
            cComment = cComment + 1
            fHeader = 1
         elseif rightstr( line, length( w2)) = w2 & cComment > 0 then
            cComment = cComment - 1
         endif
      elseif words( CommentChars) = 1 then
         if leftstr( line, 1) = CommentChars then
            cComment = 1
            fHeader = 1
         else
            leave
         endif
      else
         leave
      endif

      if fHeader then
         deleteline
         if .line = .last then
            leave
         endif
         if cComment = 0 then
            leave
         endif
      else
         if .line = .last then
            leave
         endif
         '+1'
         iterate
      endif
   enddo

   display 1
   InfolineRefresh = 1
   .autosave = saved_autosave
   if .modify > saved_modify then
      .modify = saved_modify + 1
   endif

; ---------------------------------------------------------------------------
; A common routine to parse a DOS file name.  Optional second argument
; gives source for = when used for path or fileid.
; New: Quotes, double quotes and spaces are handled correctly.
;      Duplicated '\' between path and name are avoided.
;      Environment vars are resolved.
;      ?: is replaced with the bootdrive.
;      Works now with temp files (starting with '.') as well.
;      Always returns 0.
defproc Parse_Filename( var filename)

   sourcefile = strip( arg( 2))
   if sourcefile = '' then
      if leftstr( .filename, 1) = '.' then
         sourcefile = strip( directory(), 't', '\')'\.'
      else
         sourcefile = .filename
      endif
   endif

   -- resolve every word separately
   rest = filename
   resolved = ''
   do while rest <> ''
      if leftstr( rest, 1) = "'" then
         -- resolve quoted wrd
         parse value rest with "'"wrd"'" rest
         -- parse wrd again without the quotes (wrd may contain multiple wrds)
         call Parse_Filename( wrd, sourcefile)
         resolved = resolved" '"wrd"'"
      elseif leftstr( rest, 1) = '"' then
         -- resolve double-quoted wrd
         parse value rest with '"'wrd'"' rest
         -- wrd is ready to resolve '='
         call Parse_Filename2( wrd, sourcefile)
         resolved = resolved' "'wrd'"'
      else
         -- resolve wrd
         parse value rest with wrd rest
         -- wrd is ready to resolve '='
         call Parse_Filename2( wrd, sourcefile)
         resolved = resolved' 'wrd
      endif
   enddo
   resolved = strip( resolved)
   filename = resolved
   return 0

; ---------------------------------------------------------------------------
; Resolve '=', '%' and ?: in a single word or word with spaces
defproc Parse_Filename2( var wrd, sourcefile)

   -- Parse sourcefile
   lp1 = lastpos( '\', sourcefile)
   spath = substr( sourcefile, 1, lp1)
   sname = substr( sourcefile, lp1 + 1)
   lp2 = lastpos( '.', sname)
   if lp2 > 1 then
      sbase = substr( sname, 1, lp2 - 1)
   else
      sbase = sname
   endif
   sext  = substr( sname, lp2 + 1)

   -- Replace environment variables and boot drive
   if pos( '%', wrd) |  pos( '?:', wrd) then
      wrd = ExpandEnvVars( wrd)
   endif

   -- Replace '='
   if pos( '=', wrd) then
      p = pos( '=', wrd)
      lwrd = substr( wrd, 1, p - 1)
      rwrd = substr( wrd, p + 1)
      -- base'.='     ==> base'.'sext
      -- '=.'ext      ==> sbase'.'ext
      -- '='          ==> sname
      -- '=\'name     ==> name
      -- '='name      ==> name
      -- path'\='     ==> path''sname
      -- path'='      ==> path''sname
      -- path'=.'ext  ==> path''sbase'.'ext
      -- path'\=.'ext ==> path''sbase'.'ext
      if rightstr( wrd, 2) = '.=' then
         wrd = lwrd''sext
      elseif leftstr( wrd, 2) = '=.' then
         wrd = sbase''rwrd
      elseif wrd = '=' then
         wrd = sname
      elseif leftstr( wrd, 2) = '=\' & rwrd <> '' then
         wrd = substr( rwrd, 2)
      elseif leftstr( wrd, 1) = '=' & rwrd <> '' then
         wrd = rwrd
      elseif rightstr( wrd, 2) = '\=' then
         wrd = strip( lwrd, 't', '\')'\'sname
      elseif rightstr( wrd, 2) = '=' then
         wrd = strip( lwrd, 't', '\')'\'sname
      elseif substr( wrd, p, 2) = '=.' & lwrd <> '' & rwrd <> '' then
         wrd = strip( lwrd, 't', '\')'\'sbase''rwrd
      elseif substr( wrd, p - 1, 3) = '\=.' & lwrd <> '' & rwrd <> '' then
         wrd = strip( lwrd, 't', '\')'\'sbase''rwrd
      endif
      -- If resolved wrd doesn't contain ':\' then prepend spath.
      if pos( ':\', wrd) = 0 & pos( '\\', wrd) = 0 then
         wrd = strip( spath, 't', '\')'\'wrd
      endif
   endif
   return

; ---------------------------------------------------------------------------
; Find a filemask, maybe relative, with a given startdir.
; Doesn't check if file exists.
; Doesn't resolve envvars.
; Doesn't resolve "=".
; Supports UNC and Unix names, including something like ftp://.
; Not tested with IBM host names.
; Syntax: GetFullName( FileMask [, StartDir])
; FileMask: any maybe relative dir or file filemask, wildcards allowed
; StartDir: parts of StartDir are appended to FileMask, if it's relative
defproc GetFullName( FileMask)
   FullName = FileMask

   -- Get current dir and drive
   CurDir = directory()
   CurDrive = leftstr( CurDir, 2)

   -- Get specified dir
   StartDir = arg( 2)

   -- Translate '/' to '\' if local masks specified
   p1 = pos( ':/', FileMask)
   p2 = pos( '://', FileMask)
   p3 = pos( ':/', StartDir)
   p4 = pos( '://', StartDir)
   if (p1 > 0 & p1 <> p2) | (p3 > 0 & p3 <> p4) then
      FileMask = translate( FileMask, '\', '/')
      StartDir = translate( StartDir, '\', '/')
   endif

   -- Init
   if StartDir = '' then
      StartDir = CurDir
   endif
   StartDrive = ''
   Server = ''

   -- Determine StartDir and StartDrive or Server
   if substr( StartDir, 2, 2) = ':\' then        -- fully qualified
      StartDrive = leftstr( StartDir, 2)
   elseif leftstr( StartDir, 2, 2) <> '\\' then  -- UNC
      p1 = 3
      p2 = pos( '\', StartDir, p1)
      if p2 = 0 then
         Server = StartDir
      else
         Server = leftstr( StartDir, p2 - 1)
      endif
   elseif pos( '://', StartDir ) then            -- something like ftp:// or http://
      p1 = pos( '://', StartDir ) + 3
      p2 = pos( '/', StartDir, p1)
      if p2 = 0 then
         Server = StartDir
      else
         Server = leftstr( StartDir, p2 - 1)
      endif
   elseif leftstr( StartDir, 1) = '\' then       -- local, without drive
      StartDrive = CurDrive
      StartDir = StartDrive''StartDir
   else                                          -- relative
      StartDrive = CurDrive
      StartDir = strip( CurDir, 't', '\')'\'StartDir
   endif

   -- Determine fullname                   -- fully qualified
   if substr( FileMask, 2, 2) = ':\' then        -- local, fully qualified
      FullName = FileMask
   elseif leftstr( FileMask, 2, 2) = '\\' then   -- fully qualified UNC
      FullName = FileMask
   elseif pos( '://', FileMask ) then            -- fully qualified, starting with something like ftp:// or http://
      FullName = Server''FileMask
                                           -- only drive or server missing
   elseif leftstr( FileMask, 1) = '\' &          -- UNC mask without server
      leftstr( StartDir, 2, 2) = '\\' then
      FullName = Server''FileMask
   elseif leftstr( FileMask, 1) = '/' &          -- Unix mask without server
      pos( '://', StartDir ) then
      FullName = Server''FileMask
   elseif leftstr( FileMask, 1) = '\' then       -- without drive
      FullName = StartDrive''FileMask
                                           -- relative
   elseif leftstr( StartDir, 2, 2) = '\\' then   -- relative UNC mask
      FullName = strip( StartDir, 't', '\')'\'FileMask
   elseif pos( '://', StartDir ) then            -- relative Unix mask
      FullName = strip( StartDir, 't', '/')'/'FileMask
   else                                          -- relative
      FullName = strip( StartDir, 't', '\')'\'FileMask
   endif

   -- Resolve '.' and '..' in FullName
   if not pos( '/', FullName) then
      FullName = NepmdQueryFullname( FullName)
   endif

   return FullName

; ---------------------------------------------------------------------------
; This proc is called by defc app, append, put and defc save.
; Wrong: [Does *not* assume all options are specified before filenames.]
; Options must be specified before filename.
defproc Parse_Leading_Options( var rest, var options)
   options = ''
   loop
      parse value rest with wrd more
      if substr( wrd, 1, 1) = '/' then
         options = options wrd
         rest = more
      else
         leave
      endif
   endloop

; ---------------------------------------------------------------------------
; This proc is called by defc app, append, put.
; A common routine to parse an argument string containing a mix of
; options and DOS file specs.  The DOS file specs can contain an "=" for the
; path or the fileid, which will be replaced by the corresponding part of the
; previous file (initially, the current filename).
defproc Parse_File_N_Opts( argstr)
   prev_filename = .filename
   output = ''
   do while argstr <> ''
      parse value argstr with filename rest
      if leftstr( filename, 1) = '"' then
         parse value argstr with '"' filename '"' argstr
         filename = '"'filename'"'
      else
         argstr = rest
      endif
      if substr( filename, 1, 1) <> '/' then
         call Parse_Filename( filename, prev_filename)
         prev_filename = filename
      endif
      output = output filename
   end
   return substr( output, 2)

; ---------------------------------------------------------------------------
; Called when the EPM file icon is dropped on another EPM window, while the
; source file is modified.
; We can just execute 'get' to achieve the same result.
defc Insert_Text_File
   'Get 'arg( 1)

